"""
Implementations of ABCs that use resources available to an HTCondor pool.
"""

# Recipe for performing remote submission using the Python bindings:
# https://github.com/CHTC/recipes/tree/remote-submit-htc/workflows-htc/remote-submit

import os
import time
from typing import Any

import htcondor2 as htcondor

from vertebrate.compute import Environment


class HTCondorEnvironment(Environment):
    """
    Execute jobs on an HTCondor cluster.
    """

    # NOTE: Currently hard-coded to CHTC
    SUBMIT_REQUIREMENTS = '(TARGET.PoolName == "CHTC") && ((((Target.OpSysMajorVer == 9) && (Target.OpSysName =!= "Debian") &&  !(Target.OSPool ?: false))) || (((Target.OSPool ?: false) && ((Target.OSGVO_OS_STRING ?: "") == "RHEL 9")))) && (TARGET.Arch == "X86_64") && (TARGET.OpSys == "LINUX") && (TARGET.Disk >= RequestDisk) && (TARGET.Memory >= RequestMemory) && ((TARGET.FileSystemDomain == MY.FileSystemDomain) || (TARGET.HasFileTransfer))'

    DONE = 4

    def __init__(
        self,
        user_name: str,
        ap_name: str,
        collector_name: str,
        token_dir: str = os.path.expanduser("~/.condor/tokens.d"),
        poll_secs: float = 60.0,
        debug: bool = False,
    ) -> None:
        self.user_name = user_name
        self.ap_name = ap_name
        self.collector = htcondor.Collector(collector_name)
        self.access_point = htcondor.Schedd(
            self.collector.locate(htcondor.DaemonType.Schedd, ap_name)
        )
        self._generate_creds(token_dir)
        self.poll_secs = poll_secs
        if debug:
            htcondor.enable_debug()

    def _generate_creds(self, token_dir: str) -> None:
        htcondor.param["SEC_TOKEN_DIRECTORY"] = token_dir
        credd_ad = self.collector.locate(htcondor.DaemonType.Credd, self.ap_name)
        credd = htcondor.Credd(credd_ad)
        for service in ["rdrive", "scitokens"]:
            credd.add_user_service_cred(htcondor.CredType.OAuth, b"", service)

    def _wait_for_completion(self, submit_object):
        done = False
        while not done:
            time.sleep(self.poll_secs)
            ads = self.access_point.query(
                f"ClusterID == {submit_object.cluster()}",
                projection = ["ProcID", "JobStatus"],
            )
            done = all(i['JobStatus'] == HTCondorEnvironment.DONE for i in ads)

    def execute(self, executable: str, args: tuple = (), kwargs: dict = {}) -> Any:
        kwargs["executable"] = executable
        job = htcondor.Submit(kwargs)
        submit_object = self.access_point.submit(job, spool=True)
        self.access_point.spool(submit_object)
        self._wait_for_completion(submit_object)
        self.access_point.retrieve(f"ClusterID == {submit_object.cluster()}")
        self.access_point.edit(submit_object.cluster(), "LeaveJobInQueue", False)

    def __contains__(self, item: str) -> bool:
        return os.path.exists(item)



"""
Implementation of a PostgreSQL database as an execution environment.
"""

from typing import Any

from sqlalchemy import create_engine, Engine, MetaData, Table
from sqlalchemy.sql import and_, func, functions, quoted_name, select

from vertebrate.compute import Environment


class PostgresEnvironment(Environment):
    """
    Call stored procedures defined in a PostgreSQL database.
    """

    @staticmethod
    def _postgres_engine(user: str, dbname: str, host: str, port: str) -> Engine:
        url = f"postgresql+psycopg2://{user}@{host}:{port}/{dbname}"
        return create_engine(url, echo=False)

    def __init__(self, user: str, dbname: str, host: str, port: str) -> None:
        self.engine = PostgresEnvironment._postgres_engine(user, dbname, host, port)
        self.meta = MetaData()
        self.conn = self.engine.connect()

    def __enter__(self):
        self.trans = self.conn.begin()
        return self

    def __exit__(self, type_, value, traceback):
        if type_:
            self.trans.rollback()
        else:
            self.trans.commit()

    @staticmethod
    def _identifier_name(id_name: str) -> str:
        if not id_name:
            return id_name
        if id_name.startswith('"'):
            return id_name.strip('"')
        else:
            return id_name.lower()

    @staticmethod
    def _parse_exec_name(exec_name: str) -> dict[str, str]:

        # Extract schema name from executable name if present
        schema_name = None
        if "." in exec_name:
            schema_name, exec_name = exec_name.split(".")

        return {
            "schema": PostgresEnvironment._identifier_name(schema_name),
            "name": PostgresEnvironment._identifier_name(exec_name),
        }

    def _pg_proc(self) -> Table:
        return Table(
            "pg_proc",
            self.meta,
            autoload_with=self.engine,
            schema="pg_catalog",
        )

    def _pg_namespace(self) -> Table:
        return Table(
            "pg_namespace",
            self.meta,
            autoload_with=self.engine,
            schema="pg_catalog",
        )

    def _contains_unqualified(self, exec_name: str) -> bool:
        p = self._pg_proc()
        n = self._pg_namespace()
        num_items = self.conn.execute(
            select(func.count()).\
            where(and_(
                p.c.proname == exec_name,
                p.c.pronamespace == n.c.oid,
                n.c.nspname == func.any(func.current_schemas(True)),
            ))
        ).scalar_one()
        return num_items > 0

    def _contains_qualified(self, schema_name: str, exec_name: str) -> bool:
        p = self._pg_proc()
        n = self._pg_namespace()
        num_items = self.conn.execute(
            select(func.count()).\
            where(and_(
                p.c.proname == exec_name,
                p.c.pronamespace == n.c.oid,
                n.c.nspname == schema_name,
            ))
        ).scalar_one()
        return num_items > 0

    def __contains__(self, item: str) -> bool:
        info = PostgresEnvironment._parse_exec_name(item)
        with self:
            if info["schema"] is None:
                return self._contains_unqualified(info["name"])
            else:
                return self._contains_qualified(info["schema"], info["name"])

    @staticmethod
    def _sql_call(executable: str) -> functions._FunctionGenerator:
        info = PostgresEnvironment._parse_exec_name(executable)
        if info["schema"] is None:
            sql_call = getattr(func, quoted_name(info["name"], quote=True))
        else:
            sql_schema = getattr(func, quoted_name(info["schema"], quote=True))
            sql_call = getattr(sql_schema, quoted_name(info["name"], quote=True))
        return sql_call        

    def execute(self, executable: str, args: tuple = (), kwargs: dict = {}) -> Any:
        with self:
            result = self.conn.execute(
                select(PostgresEnvironment._sql_call(executable)(*args))
            )
            return result



"""
Implementations of ABCs that use remote OS resources via SSH.
"""

import subprocess
from typing import Any

from vertebrate.compute import Environment


class SSHEnvironment(Environment):
    """
    Execute processes on a remote OS via SSH.
    """

    # These implementations are often a matter of expression;
    # SSH could be called directly using an OSEnvironment instance.

    def __init__(self, user: str, host: str) -> None:
        self.user = user
        self.host = host

    def _exec_ssh(self, executable: str, args: tuple = (), kwargs: dict = {}) -> Any:
        completed = subprocess.run(
            ("ssh", f"{self.user}@{self.host}", executable) + args,
            capture_output=True,
        )
        return completed.stdout    

    def __contains__(self, item: str) -> bool:
        exe_path = self._exec_ssh("which", args=(item,))
        return len(exe_path) > 0

    def execute(self, executable: str, args: tuple = (), kwargs: dict = {}) -> Any:
        return self._exec_ssh(executable, args=args, kwargs=kwargs)
